/*! *****************************************************************************
Copyright (c) Microsoft Corporation.

Permission to use, copy, modify, and/or distribute this software for any
purpose with or without fee is hereby granted.

THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
PERFORMANCE OF THIS SOFTWARE.
***************************************************************************** */
/* global Reflect, Promise */

var extendStatics = function(d, b) {
    extendStatics = Object.setPrototypeOf ||
        ({ __proto__: [] } instanceof Array && function (d, b) { d.__proto__ = b; }) ||
        function (d, b) { for (var p in b) if (Object.prototype.hasOwnProperty.call(b, p)) d[p] = b[p]; };
    return extendStatics(d, b);
};

function __extends(d, b) {
    if (typeof b !== "function" && b !== null)
        throw new TypeError("Class extends value " + String(b) + " is not a constructor or null");
    extendStatics(d, b);
    function __() { this.constructor = d; }
    d.prototype = b === null ? Object.create(b) : (__.prototype = b.prototype, new __());
}

function tokenize(text, parsers) {
    var tokens = [];
    while (text) {
        var offset = text.length;
        var token = void 0;
        for (var type in parsers) {
            var match = parsers[type].exec(text);
            if (match && match.index < offset) {
                var value = match[0], matches = match.slice(1);
                offset = match.index;
                token = { type: type, value: value, matches: matches };
            }
        }
        if (offset || !token) {
            throw new Error("Unable to tokenize \"" + text.slice(0, offset) + "\".");
        }
        tokens.push(token);
        text = text.slice(token.value.length);
    }
    return tokens;
}

var parsers = {
    space: /[\t ]+/,
    newline: /\r?\n/,
    colon: /:/,
    feature: /Feature/,
    scenario: /Scenario/,
    background: /Background/,
    given: /Given/,
    when: /When/,
    then: /Then/,
    and: /And/,
    but: /But/,
    word: /\w+/,
    punctuation: /./,
};
function parse(text) {
    return document(tokenize(text, parsers));
}
var ParseError = (function (_super) {
    __extends(ParseError, _super);
    function ParseError(token, message) {
        var _this = _super.call(this) || this;
        if (token) {
            var json = JSON.stringify(token.value);
            _this.message = message + ", but found " + token.type + " " + json;
        }
        else {
            _this.message = message;
        }
        return _this;
    }
    return ParseError;
}(Error));
function document(tokens) {
    var features = [];
    var token;
    while ((token = tokens.shift())) {
        if (token.type === "newline")
            continue;
        if (token.type === "space")
            continue;
        if (token.type !== "feature") {
            throw new ParseError(token, "Expected feature");
        }
        features.push(feature(tokens));
    }
    return features;
}
function feature(tokens) {
    var token;
    if (!(token = tokens.shift()) || token.type !== "colon") {
        throw new ParseError(token, "Expected a colon after the feature keyword");
    }
    var name = line(tokens);
    var description = [];
    var backgrounds = [];
    var scenarios = [];
    while ((token = tokens.shift())) {
        if (token.type === "newline")
            continue;
        if (token.type === "space")
            continue;
        if (token.type === "word") {
            description.push(token.value);
            description.push(line(tokens));
            continue;
        }
        if (token.type === "background") {
            backgrounds.push(background(tokens));
            continue;
        }
        if (token.type === "scenario") {
            scenarios.push(scenario(tokens));
            continue;
        }
        if (token.type === "feature") {
            tokens.unshift(token);
            break;
        }
        throw new ParseError(token, "Expected background or scenario");
    }
    return {
        name: name,
        description: description.join(" "),
        scenarios: scenarios,
        backgrounds: backgrounds,
    };
}
function background(tokens) {
    var token;
    if (!(token = tokens.shift()) || token.type !== "colon") {
        throw new ParseError(token, "Expected a colon after the background keyword");
    }
    var name = line(tokens);
    var description = [];
    var givens = [];
    while ((token = tokens.shift())) {
        if (token.type === "newline")
            continue;
        if (token.type === "space")
            continue;
        if (token.type === "word") {
            description.push(token.value);
            description.push(line(tokens));
            continue;
        }
        if (token.type === "given" || token.type === "and") {
            givens.push(line(tokens));
            continue;
        }
        if (token.type === "feature" || token.type === "scenario") {
            tokens.unshift(token);
            break;
        }
        throw new ParseError(token, "Expected given");
    }
    return {
        name: name,
        description: description.join(" "),
        givens: givens,
    };
}
function scenario(tokens) {
    var token;
    if (!(token = tokens.shift()) || token.type !== "colon") {
        throw new ParseError(token, "Expected a colon after the scenario keyword");
    }
    var name = line(tokens);
    var description = [];
    var givens = [];
    var whens = [];
    var thens = [];
    var group;
    while ((token = tokens.shift())) {
        if (token.type === "newline")
            continue;
        if (token.type === "space")
            continue;
        if (token.type === "word") {
            description.push(token.value);
            description.push(line(tokens));
            continue;
        }
        if (token.type === "given") {
            givens.push(line(tokens));
            group = givens;
            continue;
        }
        if (token.type === "when") {
            whens.push(line(tokens));
            group = whens;
            continue;
        }
        if (token.type === "then") {
            thens.push(line(tokens));
            group = thens;
            continue;
        }
        if (token.type === "and" || token.type === "but") {
            if (!group) {
                throw new ParseError(token, "Expected a given, when, or then");
            }
            group.push(line(tokens));
            continue;
        }
        if (token.type === "feature" || token.type === "scenario") {
            tokens.unshift(token);
            break;
        }
        throw new ParseError(token, "Expected a given, when, or then");
    }
    return {
        name: name,
        description: description.join(" "),
        givens: givens,
        whens: whens,
        thens: thens,
    };
}
function line(tokens) {
    var token;
    var line = "";
    while ((token = tokens.shift())) {
        if (token.type === "newline") {
            break;
        }
        line += token.value;
    }
    return line.trim();
}

export { ParseError, parse };
